/*
 
 Main authors:
 Anton Dries <anton.dries@cs.kuleuven.be>
 
 Copyright:
 Anton Dries, 2013
 
 This file is part of the Dominance Programming Framework, and uses Gecode.
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 */

#ifndef dp_Operator_h
#define dp_Operator_h

#include <string>
#include <gecode/int.hh>
#include <gecode/search.hh>

#include "DPSpace.h"

using namespace Gecode;
using namespace std;

class OperatorChain;	// see OperatorChain.h

/**
 Base class for all operators.
 */
class Operator {
  
private :
    
    friend class OperatorChain;
    
	Operator* input;	// operator providing input
    Operator* output;	// operator accepting output
    
public :
    
    Operator();
    
public:
	
	/**
	 Run this operator (and its predecessors).
	 
	 */
    virtual void solve();
	
    /**
	 Run this operator on the given search space.
	 
	 */
    virtual void solve(DPSpace& root);
	
	/**
	 Check whether an operator earlier in the chain already set the search order.
	 
	 */
	virtual bool canApplySearchOrder() const;
    
    /**
	 Initialize the operator (and its predecessors).
	 
	 Called at the start of solving.
	 
	 - bvCount => number of boolean variables required by successors
	 - ivCount => nubmer of integer variables required by successors
	 - skipConstraints => make lightweight space (no unnecessary constraints)
	 
	 => returns a blank search space
	 */
    virtual DPSpace& initialize(int bvCount, int ivCount, bool skipConstraints) = 0;

    
	/**
	 Finalize the operator (and its predecessors).

	 Called at end of solving.
	 */
	virtual void finalize();
    
	/**
	 Process a solution generated by the input operator.
	 
	 */
    virtual void addSolution(const DPSpace& solution);
    
	/**
	 Create a clone of this operator.
	 */
    virtual Operator* clone() const = 0;
    
	/**
	 Check whether this operator has a predecessor.
	 */
    bool hasInput() const { return input != NULL; }
    
	/**
	 Get the predecessor.
	 
	 Precondition: hasInput() == true
	 */
    virtual Operator& getInput() const {
		assert(hasInput());
		return *input;
	}
    
	/**
	 Set the predecessor.
	 
	 */
    inline void setInput(Operator& new_input) {
        input = &new_input;
    }

	/**
	 Check whether this operator has a successor.
	 */
    inline bool hasOutput() const { return output != NULL; }

	/**
	 Get the successor.
	 
	 Precondition: hasOutput() == true;
	 */
    inline Operator& getOutput() const {
		assert(hasOutput());
		return *output;
	}
    
	/** 
	 Set the successor.
	 */
    inline void setOutput(Operator& new_output) {
        output = &new_output;
        new_output.setInput(*this);
    }

	/**
	 Get the root of the operator chain.
	 */
	virtual Operator& getRoot();

	/**
	 Get the root of the operator chain (const version).
	 */
    virtual const Operator& getRoot() const;
    
};




#endif
